/*
 * (C) Copyright 2001
 * Kyle Harris, Nexus Technologies, Inc. kharris@nexus-tech.net
 *
 * (C) Copyright 2001
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <linux/byteorder/swab.h>


flash_info_t flash_info[CFG_MAX_FLASH_BANKS];	/* info for FLASH chips    */

/* Board support for 1 or 2 flash devices */
#define FLASH_PORT_WIDTH16
#undef FLASH_PORT_WIDTH32

#ifdef FLASH_PORT_WIDTH16
#define FLASH_PORT_WIDTH		ushort
#define FLASH_PORT_WIDTHV		vu_short
#define SWAP(x)               __swab16(x)
#else
#define FLASH_PORT_WIDTH		ulong
#define FLASH_PORT_WIDTHV		vu_long
#define SWAP(x)               __swab32(x)
#endif

#define FPW    FLASH_PORT_WIDTH
#define FPWV   FLASH_PORT_WIDTHV

#define mb() __asm__ __volatile__ ("" : : : "memory")

/*-----------------------------------------------------------------------
 * Functions
 */
static ulong flash_get_type (FPW *addr, flash_info_t *info);
static int write_data (flash_info_t *info, ulong dest, FPW data);
static void flash_get_offsets (ulong base, flash_info_t *info);
void inline spin_wheel (void);

/*-----------------------------------------------------------------------
 */

#if 0
/* for now, we don't support Flash, so we have only dummies here */
unsigned long flash_init (void)
{
    return 0; /* for now, we don't support Flash */
}
void flash_print_info (flash_info_t *info)
{
}
int flash_erase (flash_info_t *info, int s_first, int s_last)
{
    return 0;
}
int write_buff (flash_info_t *info, uchar *src, ulong addr, ulong cnt)
{
    return 0;
}
#else

int flash_unlock( FPWV *addr)
{
	FPW status;

	*addr = (FPW) 0x00600060;       /* unlock block */
	*addr = (FPW) 0x00d000d0;

	while (((status = *addr) & (FPW) 0x00800080) != (FPW) 0x00800080) {
                if (get_timer_masked() > CFG_FLASH_ERASE_TOUT) {
			*addr = (FPW) 0x00B000B0;	/* suspend erase     */
			*addr = (FPW) 0x00FF00FF;	/* reset to read mode */

                    return 1;
                }
        }

        /* clear status register command & reset to read mode */
	*addr = (FPW) 0x00500050;	/* clear status register cmd.   */
	*addr = (FPW) 0x00FF00FF;	/* resest to read mode          */

        return 0;
}

unsigned long flash_init (void)
{
	int i;
	ulong size = 0;

	for (i = 0; i < CFG_MAX_FLASH_BANKS; i++) {
		flash_info_t *info = &flash_info[i];
		switch (i) {
		case 0:
			flash_get_type ((FPW *) PHYS_FLASH_1, info);
			flash_get_offsets (PHYS_FLASH_1, info);
			break;
		case 1:
			flash_get_type ((FPW *) PHYS_FLASH_2, info);
			flash_get_offsets (PHYS_FLASH_2, info);
			break;
		default:
			panic ("configured too many flash banks!\n");
			break;
		}
		size += flash_info[i].size;

		// unlock flash
		for (i = 0; i < info->sector_count; ++i)
			flash_unlock( (FPWV *) info->start[i]);
	}
	
	return size;
}

/*-----------------------------------------------------------------------
 */
void flash_print_info (flash_info_t *info)
{
	int i;

	if (info->flash_id == FLASH_UNKNOWN) {
		printf ("missing or unknown FLASH type\n");
		return;
	}

	switch (info->flash_id & FLASH_VENDMASK) {
	case FLASH_MAN_INTEL:
		printf ("INTEL "); break;
	default:
		printf ("Unknown Vendor(%x) ", info->flash_id );
		break;
	}

	switch (info->flash_id & FLASH_TYPEMASK) {
	case FLASH_28F128J3A:
		printf ("28F128J3A\n"); break;
	case FLASH_48F2000ZB:
		printf ("48F2000P0ZBQ0\n"); break;
	case FLASH_48F2000ZT:
		printf ("48F2000P0ZTQ0\n"); break;
	case FLASH_48F3000ZB:
		printf ("48F3000P0ZBQ0\n"); break;
	case FLASH_48F3000ZT:
		printf ("48F3000P0ZTQ0\n"); break;
	case FLASH_48F4x00ZB:
		printf ("48F4x00P0ZBQ0\n"); break;
	case FLASH_48F4x00ZT:
		printf ("48F4x00P0ZTQ0\n"); break;
	default:
		printf ("Unknown Chip Type: 0x%x\n",
			(info->flash_id & FLASH_TYPEMASK) );
		break;
	}

	printf ("  Size: %ld MB in %d Sectors\n",
			info->size >> 20, info->sector_count);

	printf ("  Sector Start Addresses:");
	for (i = 0; i < info->sector_count; ++i) {
		if ((i % 4) == 0)
			printf ("\n\t");
		printf (" %08lX\t", info->start[i]);
	}
	printf ("\n");
}

/*-----------------------------------------------------------------------
 */

FPW cfi_read_word( FPW* addr)
{
	FPW result;
	volatile FPW value;

	mb ();
	value = *addr++;
	result = (value & 0xff);
	
	mb ();
	value = *addr++;
	result |=  ((ulong)value << 8);

	return result;
}

static void flash_get_offsets ( ulong base, flash_info_t *info)
{
	FPW *addr = (FPW *)base;

	if (info->flash_id == FLASH_UNKNOWN) {
		return;
	}
	
	if ((info->flash_id & FLASH_VENDMASK) == FLASH_MAN_INTEL)
	{
		int i,j;
		FPW *cfi = addr + 0x2c;
		volatile FPW rgs;
		
		/* read CFI*/
		addr[0] = (FPW) 0x00980098;
	
		mb ();
		rgs = *cfi++;

		info->size = 0;
		info->sector_count = 0;

		for( i=0; i<rgs; i++, cfi+=4)
		{
			int rcnt = cfi_read_word(cfi) + 1;
			ulong rsz = cfi_read_word(cfi+2) << 8;

			for( j=0; j<rcnt; j++, base+=rsz)
			{
				info->start[info->sector_count] = base;
				info->sector_count++;
				info->size += rsz;
			}
		}
		addr[0] = (FPW) 0x00FF00FF;		/* restore read mode */
	}
}

/*
 * The following code cannot be run from FLASH!
 */
static ulong flash_get_type (FPW *addr, flash_info_t *info)
{
	volatile FPW value;

	info->flash_id = FLASH_UNKNOWN;
	info->sector_count = 0;
	info->size = 0;

	/* read manufacturer ID */
	*addr = (FPW) 0x00900090;
	
	mb ();
	value = addr[0];

	switch (value) {

	case (FPW) INTEL_MANUFACT:
	{
		mb ();
		value = addr[1];			/* device ID        */

		info->flash_id = FLASH_MAN_INTEL | (value & FLASH_TYPEMASK);
	} break;

	default:
		break;
	}
	*addr = (FPW) 0x00FF00FF;		/* restore read mode */

	return (info->flash_id);
}

/*-----------------------------------------------------------------------
 */

int flash_erase (flash_info_t *info, int s_first, int s_last)
{
	int flag, sect;
	ulong type, start, last;
	int rcode = 0;

	if ((s_first < 0) || (s_first > s_last)) {
		if (info->flash_id == FLASH_UNKNOWN) {
			printf ("- missing\n");
		} else {
			printf ("- no sectors to erase\n");
		}
		return 1;
	}

	type = (info->flash_id & FLASH_VENDMASK);
	if ((type != FLASH_MAN_INTEL)) {
		printf ("Can't erase unknown flash type %08lx - aborted\n",
			info->flash_id);
		return 1;
	}
	printf ("\n");
	
	start = get_timer (0);
	last = start;

	/* Disable interrupts which might cause a timeout here */
	flag = disable_interrupts ();

	/* Start erase on unprotected sectors */
	for (sect = s_first; sect <= s_last; sect++) {
		FPWV *addr = (FPWV *) (info->start[sect]);
		FPW status;

		printf ("Erasing sector %2d ... ", sect);

		/* arm simple, non interrupt dependent timer */
		reset_timer_masked ();

		if( flash_unlock(addr))
			return 1;

		*addr = (FPW) 0x00500050;	/* clear status register */
		*addr = (FPW) 0x00200020;	/* erase setup */
		*addr = (FPW) 0x00D000D0;	/* erase confirm */

		while (((status = *addr) & (FPW) 0x00800080) != (FPW) 0x00800080) {
			if (get_timer_masked () > CFG_FLASH_ERASE_TOUT) {
				printf ("Timeout\n");
				*addr = (FPW) 0x00B000B0;	/* suspend erase     */
				*addr = (FPW) 0x00FF00FF;	/* reset to read mode */
				rcode = 1;
				break;
			}
		}

		*addr = (FPW) 0x00500050;	/* clear status register cmd.   */
		*addr = (FPW) 0x00FF00FF;	/* resest to read mode          */

		printf (" done\n");
	}
	return rcode;
}

/*-----------------------------------------------------------------------
 * Copy memory to flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 * 4 - Flash not identified
 */

int write_buff (flash_info_t *info, uchar *src, ulong addr, ulong cnt)
{
	ulong cp, wp;
	FPW data;
	int count, i, l, rc, port_width;

	if (info->flash_id == FLASH_UNKNOWN) {
		return 4;
	}
/* get lower word aligned address */
#ifdef FLASH_PORT_WIDTH16
	wp = (addr & ~1);
	port_width = 2;
#else
	wp = (addr & ~3);
	port_width = 4;
#endif
	/*
	 * handle unaligned start bytes
	 */
	if ((l = addr - wp) != 0) {
		data = 0;
		for (i = 0, cp = wp; i < l; ++i, ++cp) {
			data = (data << 8) | (*(uchar *) cp);
		}
		for (; i < port_width && cnt > 0; ++i) {
			data = (data << 8) | *src++;
			--cnt;
			++cp;
		}
		for (; cnt == 0 && i < port_width; ++i, ++cp) {
			data = (data << 8) | (*(uchar *) cp);
		}

		if ((rc = write_data (info, wp, SWAP (data))) != 0) {
			return (rc);
		}
		wp += port_width;
	}

	/*
	 * handle word aligned part
	 */
	count = 0;
	while (cnt >= port_width) {
		data = 0;
		for (i = 0; i < port_width; ++i) {
			data = (data << 8) | *src++;
		}
		if ((rc = write_data (info, wp, SWAP (data))) != 0) {
			return (rc);
		}
		wp += port_width;
		cnt -= port_width;
		if (count++ > 0x800) {
			spin_wheel ();
			count = 0;
		}
	}

	if (cnt == 0) {
		return (0);
	}

	/*
	 * handle unaligned tail bytes
	 */
	data = 0;
	for (i = 0, cp = wp; i < port_width && cnt > 0; ++i, ++cp) {
		data = (data << 8) | *src++;
		--cnt;
	}
	for (; i < port_width; ++i, ++cp) {
		data = (data << 8) | (*(uchar *) cp);
	}

	return (write_data (info, wp, SWAP (data)));
}

/*-----------------------------------------------------------------------
 * Write a word or halfword to Flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 */
static int write_data (flash_info_t *info, ulong dest, FPW data)
{
	FPWV *addr = (FPWV *) dest;
	ulong status;
	int flag;

	/* Check if Flash is (sufficiently) erased */
	if ((*addr & data) != data) {
		printf ("not erased at %08lx (%lx)\n", (ulong) addr, *addr);
		return (2);
	}
	/* Disable interrupts which might cause a timeout here */
	flag = disable_interrupts ();

	*addr = (FPW) 0x00400040;	/* write setup */
	*addr = data;

	/* arm simple, non interrupt dependent timer */
	reset_timer_masked ();

	/* wait while polling the status register */
	while (((status = *addr) & (FPW) 0x00800080) != (FPW) 0x00800080) {
		if (get_timer_masked () > CFG_FLASH_WRITE_TOUT) {
			*addr = (FPW) 0x00FF00FF;	/* restore read mode */
			return (1);
		}
	}

	*addr = (FPW) 0x00FF00FF;	/* restore read mode */

	return (0);
}

void inline spin_wheel (void)
{
	static int p = 0;
	static char w[] = "\\/-";

	printf ("\010%c", w[p]);
	(++p >= 3) ? (p = 0) : 0;
}

#endif /* 0 */
