/*
 * (C) Copyright 2005
 * Detlef Vollmann, Toradex AG, <dv@vollmann.ch>
 *
 * (C) Copyright 2002
 * Wolfgang Denk, DENX Software Engineering, <wd@denx.de>
 *
 * (C) Copyright 2002
 * Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Marius Groeger <mgroeger@sysgo.de>
 *
 * (C) Copyright 2002
 * Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Alex Zuepke <azu@sysgo.de>
 *
 * Copyright (C) 1999 2000 2001 Erik Mouw (J.A.K.Mouw@its.tudelft.nl)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <common.h>
//#include <watchdog.h>
#include <asm/arch/pxa-regs.h>

#define PXA_BAUD_CLOCK 921600

static inline unsigned int baud2quot(unsigned int baudrate)
{
    return PXA_BAUD_CLOCK / baudrate;
}

#if defined(CONFIG_FFUART) || (defined(CONFIG_KGDB_SER_INDEX) && CONFIG_KGDB_SER_INDEX == 1)
static void ffuart_setup(unsigned int quot)
{
	CKEN |= CKEN6_FFUART;

	FFIER = 0;					/* Disable for now */
	FFFCR = 0;					/* No fifos enabled */

	/* set baud rate */
	FFLCR = LCR_WLS0 | LCR_WLS1 | LCR_DLAB;
	FFDLL = quot & 0xff;
	FFDLH = quot >> 8;
	FFLCR = LCR_WLS0 | LCR_WLS1;

	FFIER = IER_UUE;			/* Enable FFUART */
}
#endif

#if defined(CONFIG_BTUART) || (defined(CONFIG_KGDB_SER_INDEX) && CONFIG_KGDB_SER_INDEX == 2)
static btuart_setup(unsigned int quot)
{
	CKEN |= CKEN7_BTUART;

	BTIER = 0;
	BTFCR = 0;

	/* set baud rate */
	BTLCR = LCR_DLAB;
	BTDLL = quot & 0xff;
	BTDLH = quot >> 8;
	BTLCR = LCR_WLS0 | LCR_WLS1;

	BTIER = IER_UUE;			/* Enable BTUART */
}
#endif

#if defined(CONFIG_STUART) || (defined(CONFIG_KGDB_SER_INDEX) && CONFIG_KGDB_SER_INDEX == 3)
static stuart_setup(unsigned int quot)
{
	CKEN |= CKEN5_STUART;

	STIER = 0;
	STFCR = 0;

	/* set baud rate */
	STLCR = LCR_DLAB;
	STDLL = quot & 0xff;
	STDLH = quot >> 8;
	STLCR = LCR_WLS0 | LCR_WLS1;

	STIER = IER_UUE;			/* Enable STUART */
}
#endif

void serial_setbrg (void)
{
	DECLARE_GLOBAL_DATA_PTR;

#ifdef CONFIG_FFUART
    ffuart_setup(baud2quot(gd->baudrate));

#elif defined(CONFIG_BTUART)
    btuart_setup(baud2quot(gd->baudrate));

#elif defined(CONFIG_STUART)
    stuart_setup(baud2quot(gd->baudrate));

#else
#error "Bad: you didn't configure serial ..."
#endif
}


/*
 * Initialise the serial port with the given baudrate. The settings
 * are always 8 data bits, no parity, 1 stop bit, no start bits.
 *
 */
int serial_init (void)
{
	serial_setbrg ();

	return (0);
}


/*
 * Output a single byte to the serial port.
 */
void serial_putc (const char c)
{
#ifdef CONFIG_FFUART
	/* wait for room in the tx FIFO on FFUART */
	while ((FFLSR & LSR_TEMT) == 0);
	FFTHR = c;
#elif defined(CONFIG_BTUART)
	while ((BTLSR & LSR_TEMT ) == 0 );
	BTTHR = c;
#elif defined(CONFIG_STUART)
	while ((STLSR & LSR_TEMT ) == 0 );
	STTHR = c;
#endif

	/* If \n, also do \r */
	if (c == '\n')
		serial_putc ('\r');
}

/*
 * Read a single byte from the serial port. Returns 1 on success, 0
 * otherwise. When the function is succesfull, the character read is
 * written into its argument c.
 */
int serial_tstc (void)
{
#ifdef CONFIG_FFUART
	return FFLSR & LSR_DR;
#elif defined(CONFIG_BTUART)
	return BTLSR & LSR_DR;
#elif defined(CONFIG_STUART)
	return STLSR & LSR_DR;
#endif
}

/*
 * Read a single byte from the serial port. Returns 1 on success, 0
 * otherwise. When the function is succesfull, the character read is
 * written into its argument c.
 */
int serial_getc (void)
{
#ifdef CONFIG_FFUART
	while (!(FFLSR & LSR_DR));
	return (char) FFRBR & 0xff;
#elif defined(CONFIG_BTUART)
	while (!(BTLSR & LSR_DR));
	return (char) BTRBR & 0xff;
#elif defined(CONFIG_STUART)
	while (!(STLSR & LSR_DR));
	return (char) STRBR & 0xff;
#endif
}

void
serial_puts (const char *s)
{
	while (*s) {
		serial_putc (*s++);
	}
}

#ifdef CONFIG_KGDB_SER_INDEX
int kgdb_serial_init(void)
{
#if CONFIG_KGDB_SER_INDEX == 1
    ffuart_setup(baud2quot(CONFIG_KGDB_BAUDRATE));

#elif CONFIG_KGDB_SER_INDEX == 2
    btuart_setup(baud2quot(CONFIG_KGDB_BAUDRATE));

#elif CONFIG_KGDB_SER_INDEX == 3
    stuart_setup(baud2quot(CONFIG_KGDB_BAUDRATE));

#else
#error "Bad: you didn't configure serial for KGDB..."
#endif
}

void putDebugChar(const char c)
{
#if CONFIG_KGDB_SER_INDEX == 1
	/* wait for room in the tx FIFO on FFUART */
	while ((FFLSR & LSR_TEMT) == 0)
		WATCHDOG_RESET ();	/* Reset HW Watchdog, if needed */
	FFTHR = c;
#elif CONFIG_KGDB_SER_INDEX == 2
	while ((BTLSR & LSR_TEMT ) == 0 )
		WATCHDOG_RESET ();	/* Reset HW Watchdog, if needed */
	BTTHR = c;
#elif CONFIG_KGDB_SER_INDEX == 3
	while ((STLSR & LSR_TEMT ) == 0 )
		WATCHDOG_RESET ();	/* Reset HW Watchdog, if needed */
	STTHR = c;
#endif

	/* If \n, also do \r */
	if (c == '\n')
		serial_putc ('\r');
}

int getDebugChar(void)
{
#ifdef CONFIG_KGDB_SER_INDEX ==13
	while (!(FFLSR & LSR_DR))
		WATCHDOG_RESET ();	/* Reset HW Watchdog, if needed */
	return (char) FFRBR & 0xff;
#elif CONFIG_KGDB_SER_INDEX == 2
	while (!(BTLSR & LSR_DR))
		WATCHDOG_RESET ();	/* Reset HW Watchdog, if needed */
	return (char) BTRBR & 0xff;
#elif CONFIG_KGDB_SER_INDEX == 3
	while (!(STLSR & LSR_DR))
		WATCHDOG_RESET ();	/* Reset HW Watchdog, if needed */
	return (char) STRBR & 0xff;
#endif
}
#endif
